///////////////////////////////////////////////
//
// A L-System fractal generator
// Copyright (c) 2005 Jari Komppa
//
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
//     This software is provided 'as-is', without any express or implied
//     warranty.    In no event will the authors be held liable for any damages
//     arising from the use of this software.
// 
//     Permission is granted to anyone to use this software for any purpose,
//     including commercial applications, and to alter it and redistribute it
//     freely, subject to the following restrictions:
// 
//     1. The origin of this software must not be misrepresented; you must not
//        claim that you wrote the original software. If you use this software
//        in a product, an acknowledgment in the product documentation would be
//        appreciated but is not required.
//     2. Altered source versions must be plainly marked as such, and must not be
//        misrepresented as being the original software.
//     3. This notice may not be removed or altered from any source distribution.
// 
// (eg. same as ZLIB license)
// 
//
///////////////////////////////////////////////

#ifdef _MSC_VER
#include <windows.h> // needed to get GL stuff to work
#endif
#include <SDL.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lsystem.h"
#include "tokenwad.h"

#define PI 3.1415926535897932384626433832795f


void lsystem::push(int aWhat, vertex &aPos, vertex &aDir, float aLen, int aAxis, int aColor)
{
    mStack[mStackptr].mWhat = aWhat;
    if (aWhat & STACK_POS) mStack[mStackptr].mPos = aPos;
    if (aWhat & STACK_DIR) mStack[mStackptr].mDir = aDir;
    if (aWhat & STACK_LEN) mStack[mStackptr].mLen = aLen;
    if (aWhat & STACK_AXIS) mStack[mStackptr].mAxis = aAxis;
    if (aWhat & STACK_COLOR) mStack[mStackptr].mColor = aColor;
    mStackptr++;
}

int lsystem::pop(vertex &aPos, vertex &aDir, float &aLen, int &aAxis, int &aColor)
{
    mStackptr--;
    if (mStack[mStackptr].mWhat & STACK_POS) aPos = mStack[mStackptr].mPos;
    if (mStack[mStackptr].mWhat & STACK_DIR) aDir = mStack[mStackptr].mDir;
    if (mStack[mStackptr].mWhat & STACK_LEN) aLen = mStack[mStackptr].mLen;
    if (mStack[mStackptr].mWhat & STACK_AXIS) aAxis = mStack[mStackptr].mAxis;
    if (mStack[mStackptr].mWhat & STACK_COLOR) aColor = mStack[mStackptr].mColor;
    return mStack[mStackptr].mWhat;
}

float lsystem::tick()
{
	return (float)SDL_GetTicks();
}

lsystem::lsystem()
{
    mAngle = 30 / 180.0f * 3.14f;
    mWave = 0;
    mLength = 1;
    mGenlength = 1;
    mStyle = 0;
    mAxiomcount = 0;
    mIterations = 1;
    mAxiom = NULL;
    mColor[0] = 0x000040;
    mColor[1] = 0x000080;
    mColor[2] = 0x002080;
    mColor[3] = 0x004080;
    mColor[4] = 0x008080;
    mColor[5] = 0x208080;
    mColor[6] = 0x408080;
    mColor[7] = 0x808080;
    mStackptr = 0;
    int i;
    for (i = 0; i < 8; i++)
    {
        mVtxcount[i] = 0;
        mVtx[i] = 0;
    }
}

lsystem::~lsystem()
{
    int i;
    for (i = 0; i < mAxiomcount*2; i++)
        delete[] mAxiom[i];
    delete[] mAxiom;
}

void lsystem::generate(char *aBuffer)
{
	char * walker = aBuffer;
	int steps[8] = {0,0,0,0,0,0,0,0};
    int pos[8] = {0,0,0,0,0,0,0,0};
    float length = mLength;
    float rot[3];
    rot[0] = 0;
    rot[1] = 0;
    rot[2] = 0;
    walker = aBuffer;
    vertex v;
    int current_rot = 0;
    int current_vtxset = 0;
	float wave = (float)(sin(tick()*0.002) * mWave);
    
	while (*walker)
	{
        switch(*walker)
        {
            case '(':
            {
                walker++;
                int what = 0;
                switch (*walker)
                {
                    case 'c':
                    case 'a':
                        what = STACK_COLOR;
                        break;
                }
                vertex r;
                push(what, v, r, length, current_rot, current_vtxset);
            }
                break;
            case ')':
            {
                vertex r;
                pop(v, r, length, current_rot, current_vtxset);
            }
                break;
            case '1':
                current_vtxset = 0;
                break;
            case '2':
                current_vtxset = 1;
                break;
            case '3':
                current_vtxset = 2;
                break;
            case '4':
                current_vtxset = 3;
                break;
            case '5':
                current_vtxset = 4;
                break;
            case '6':
                current_vtxset = 5;
                break;
            case '7':
                current_vtxset = 6;
                break;
            case '8':
                current_vtxset = 7;
                break;
            case '#':
                current_vtxset++;
                if (current_vtxset >= 8) current_vtxset = 0;
                    break;
            case '.':
                steps[current_vtxset]++;
                break;
                
        }
		walker++;
	}
	walker = aBuffer;
    
    int i;
    for (i = 0; i < 8; i++)
    {
        delete[] mVtx[i];
        mVtxcount[i] = steps[i]*2;
        mVtx[i] = new float[mVtxcount[i]*3];
    }
    
    length = mLength;
    rot[0] = 0;
    rot[1] = 0;
    rot[2] = 0;
    walker = aBuffer;
    current_rot = 0;
    current_vtxset = 0;
    while (*walker)
    {
        switch(*walker)
        {
            case '0':
                v.mX = 0;
                v.mY = 0;
                v.mZ = 0;
                break;
            case '1':
                current_vtxset = 0;
                break;
            case '2':
                current_vtxset = 1;
                break;
            case '3':
                current_vtxset = 2;
                break;
            case '4':
                current_vtxset = 3;
                break;
            case '5':
                current_vtxset = 4;
                break;
            case '6':
                current_vtxset = 5;
                break;
            case '7':
                current_vtxset = 6;
                break;
            case '8':
                current_vtxset = 7;
                break;
            case '#':
                current_vtxset++;
                if (current_vtxset >= 8) current_vtxset = 0;
                    break;
            case '+':
                rot[current_rot] += mAngle+wave;
                break;
            case '-':
                rot[current_rot] -= mAngle+wave;
                break;
            case '/':
                current_rot++;
                if (current_rot >= 3) current_rot = 0;
                    break;
            case '\\':
                current_rot--;
                if (current_rot <= 0) current_rot = 2;
                    break;
            case 'x':
                current_rot = 0;
                break;
            case 'y':
                current_rot = 1;
                break;
            case 'z':
                current_rot = 2;
                break;
            case '(':
            {
                walker++;
                int what = 0;
                switch (*walker)
                {
                    case 'a':
                        what = STACK_EVERYTHING;
                        break;
                    case 'p':
                        what = STACK_POS;
                        break;
                    case 'd':
                        what = STACK_DIR;
                        break;
                    case 'l':
                        what = STACK_LEN;
                        break;
                    case 'x':
                        what = STACK_AXIS;
                        break;
                    case 'c':
                        what = STACK_COLOR;
                        break;
                }
                vertex r;
                r.mX = rot[0];
                r.mY = rot[1];
                r.mZ = rot[2];
                push(what, v, r, length, current_rot, current_vtxset);
            }
                break;
            case ')':
            {
                vertex r;
                int what = pop(v, r, length, current_rot, current_vtxset);
                if (what & STACK_DIR)
                {
                    rot[0] = r.mX;
                    rot[1] = r.mY;
                    rot[2] = r.mZ;
                }
            }
                break;
            case'%':
                length *= mGenlength;
                break;
            case '.':
            {
                vertex t;
                vertex u;
                t.mX = -sin(rot[2]);
                t.mY =  cos(rot[2]);
                u.mX =  t.mX * cos(rot[1]);
                t.mZ = -t.mX * sin(rot[1]);
                u.mY =  t.mY * cos(rot[0]) - t.mZ * sin(rot[0]);
                u.mZ =  t.mY * sin(rot[0]) + t.mZ * cos(rot[0]);
                mVtx[current_vtxset][pos[current_vtxset]] = v.mX; pos[current_vtxset]++;
                mVtx[current_vtxset][pos[current_vtxset]] = v.mY; pos[current_vtxset]++;
                mVtx[current_vtxset][pos[current_vtxset]] = v.mZ; pos[current_vtxset]++;
                v.mX += u.mX * length;
                v.mY += u.mY * length;
                v.mZ += u.mZ * length;
                mVtx[current_vtxset][pos[current_vtxset]] = v.mX; pos[current_vtxset]++;
                mVtx[current_vtxset][pos[current_vtxset]] = v.mY; pos[current_vtxset]++;
                mVtx[current_vtxset][pos[current_vtxset]] = v.mZ; pos[current_vtxset]++;
            }   
                break;
            case ':':
            {
                vertex t;
                vertex u;
                t.mX = -sin(rot[2]);
                t.mY =  cos(rot[2]);
                u.mX =  t.mX * cos(rot[1]);
                t.mZ = -t.mX * sin(rot[1]);
                u.mY =  t.mY * cos(rot[0]) - t.mZ * sin(rot[0]);
                u.mZ =  t.mY * sin(rot[0]) + t.mZ * cos(rot[0]);
                v.mX += u.mX * length;
                v.mY += u.mY * length;
                v.mZ += u.mZ * length;
            }   
                break;
                
        }
        walker++;
    }
    
}

void lsystem::parse(char * aTarget, char * aTemp, int aTargetlen, int aOutputDebugLog)
{
	int i = 0;
    FILE * f = NULL;
    if (aOutputDebugLog)
    {
        f = fopen("lsys.log","wb");
        fprintf(f,"L-system debug log\n\nAxioms:\n");
        i = 0;
        while (i < mAxiomcount)
        {
            fprintf(f, "%s = \"%s\"\n",mAxiom[i*2+0],mAxiom[i*2+1]);
            i++;
        }
        fprintf(f, "\n");
    }
    int lencheck = 0;
    i = 0;
	while (mAxiom[1][i])
	{
		aTemp[i] = mAxiom[1][i];
		i++;
	}
	aTemp[i] = 0;
    
	int iters = 0;
	while (iters < mIterations)
	{
        if (aOutputDebugLog)
        {
            fprintf(f,"Iteration %d:\n\"%s\"\n\n",iters,aTemp);
        }
        char * target = aTarget;
		char * walker = aTemp;
		while (*walker)
		{
			i = 0;
			while (i < mAxiomcount && *walker != mAxiom[i*2][0])
				i++;
            
			if (i < mAxiomcount)
			{
				char * d = mAxiom[i*2+1];
				while (*d)
				{
					*target = *d;
					target++;
					d++;
				}
				walker++;
                lencheck++;
                if (lencheck == aTargetlen)
                    return;
			}
			else
			{
				*target = *walker;
				target++;
				walker++;
                lencheck++;
                if (lencheck == aTargetlen)
                    return;
			}
		}
		*target = 0;
		i = 0;
		while (aTarget[i])
		{
			aTemp[i] = aTarget[i];
			i++;
            if (i == aTargetlen)
                return;
		}
		aTemp[i] = 0;
		iters++;
	}
    if (aOutputDebugLog)
    {
        fprintf(f,"Iteration %d:\n\"%s\"\n\n",iters,aTemp);
    }
        
	char * target = aTarget;
	char * walker = aTemp;
    
	while (*walker)
	{
		i = 0;
		while (i < mAxiomcount && *walker != mAxiom[i*2][0])
			i++;
		if (i >= mAxiomcount)
		{
			*target = *walker;
			target++;
		}
		walker++;
	}
	*target = 0;
    if (aOutputDebugLog)
    {
        fprintf(f,"Opcode string:\n\"%s\"\n\n",aTarget);
        fclose(f);
    }
}

void lsystem::draw()
{
    char * tgt = new char[65536];
    char * tmp = new char[65536];
	parse(tgt, tmp, 65536);
    generate(tgt);
    delete[] tgt;
    delete[] tmp;
    
    int grow = mStyle & 1;
    int style = mStyle / 2;

    glEnableClientState(GL_VERTEX_ARRAY);

    
    if (style == 0)
    {
        glDisable(GL_CULL_FACE);
        glEnable(GL_LINE_SMOOTH);
        glEnable(GL_BLEND);
        
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LEQUAL);
        
        glBlendFunc(GL_ONE, GL_ONE);
        glColor4f(0.0f,0.0f,0.0f,0.0f);
        glEnable(GL_VERTEX_ARRAY);
        glLineWidth(7);
        
        int j;
        for (j = 0; j < 8; j++)
        {
            
            glVertexPointer(3, GL_FLOAT, 0, mVtx[j]);
            
            int i = 0;
            
            if (grow)
            {
                i = (int)tick()/25;
                if (i > mVtxcount[j])
                    i = mVtxcount[j];    
            }
            else
            {
                i = mVtxcount[j];
            }
            
            glDrawArrays(GL_LINES, 0, i);
        }
        
        glDepthFunc(GL_LEQUAL);
        glEnable(GL_POLYGON_OFFSET_LINE);
        glPolygonOffset(1, 10);
        glLineWidth(2);
        glBlendFunc(GL_ONE, GL_ZERO);
        
        for (j = 0; j < 8; j++)
        {
            glColor4f(((mColor[j] >>16) & 0xff) / 255.0f,
                      ((mColor[j] >> 8) & 0xff) / 255.0f,
                      ((mColor[j] >> 0) & 0xff) / 255.0f,
                      1.0f);
            
            glVertexPointer(3, GL_FLOAT, 0, mVtx[j]);
            
            int i = 0;
            
            if (grow)
            {
                i = (int)tick()/25;
                if (i > mVtxcount[j])
                    i = mVtxcount[j];    
            }
            else
            {
                i = mVtxcount[j];
            }
            
            glDrawArrays(GL_LINES, 0, i);
        }
        
        glPolygonOffset(0,0);
    }
    
    if (style == 1)
    {
        glDisable(GL_CULL_FACE);
        glEnable(GL_LINE_SMOOTH);
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE);
        
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_POLYGON_OFFSET_LINE);
        
        int j;
        for (j = 0; j < 8; j++)
        {
            glColor4f(((mColor[j] >>16) & 0xff) / 255.0f,
                      ((mColor[j] >> 8) & 0xff) / 255.0f,
                      ((mColor[j] >> 0) & 0xff) / 255.0f,
                      1.0f);
            
            glVertexPointer(3, GL_FLOAT, 0, mVtx[j]);
            
            int i = 0;
            
            if (grow)
            {
                i = (int)tick()/25;
                if (i > mVtxcount[j])
                    i = mVtxcount[j];    
            }
            else
            {
                i = mVtxcount[j];
            }
            
            glLineWidth(5);
            glDrawArrays(GL_LINES, 0, i);
            glLineWidth(1);
            glDrawArrays(GL_LINES, 0, i);
        }
    }
}

lsystem * lsystem::load_scene(char * aFilename)
{
    lsystem *ls = new lsystem;
    
    FILE * f = fopen(aFilename, "rb");
    if (f==NULL)
        exit(-7);
    fseek(f, 0, SEEK_END);
    int flen = ftell(f);
    fseek(f, 0, SEEK_SET);
    char *d = new char[flen + 1];
    fread(d, flen, 1, f);
    fclose(f);
    d[flen]=0;
    tokenwad lines;
    lines.tokenize(d,"\r\n");
    int i;
    for (i = 0; i < lines.tokens(); i++)
    {
        if (lines.token(i)[0] != 0 && lines.token(i)[0] != ';')
        {
            tokenwad l;
            l.tokenize(lines.token(i),"=");
            if (strlen(l.token(0)) == 1)
                ls->mAxiomcount++;
        }
    }
    ls->mAxiom = new char*[ls->mAxiomcount*2]; 
    int axiomcount = 0;
    for (i = 0; i < lines.tokens(); i++)
    {
        if (lines.token(i)[0] != 0 && lines.token(i)[0] != ';')
        {
            tokenwad l;
            l.tokenize(lines.token(i),"=,");
            if (l.tokenEqualsNocase(0,"angle"))
            {
                ls->mAngle = (float)(atof(l.token(1))/180.0f*PI);
            }
            else
                if (l.tokenEqualsNocase(0,"wave"))
                {
                    ls->mWave = (float)(atof(l.token(1))/180.0f*PI);
                }
            else
                if (l.tokenEqualsNocase(0,"length"))
                {
                    ls->mLength = (float)(atof(l.token(1)));
                }
            else
                if (l.tokenEqualsNocase(0,"iterations"))
                {
                    ls->mIterations = atoi(l.token(1));
                }
            else
                if (l.tokenEqualsNocase(0,"style"))
                {
                    ls->mStyle = atoi(l.token(1));
                }
            else
                if (l.tokenEqualsNocase(0,"genlength"))
                {
                    ls->mGenlength = (float)(atof(l.token(1)));
                }
            else
                if (l.tokenEqualsNocase(0,"color"))
                {
                    int colors = (l.tokens()-1) / 3;
                    int i;
                    for (i = 0; i < colors; i++)
                    {
                        ls->mColor[i] = (((int)(atof(l.token(i*3+0+1))*0xff))<<16) |
                        (((int)(atof(l.token(i*3+1+1))*0xff))<<8) |
                        (((int)(atof(l.token(i*3+2+1))*0xff))<<0);
                    }
                }
            else
            {
                if (strlen(l.token(0)) != 1)
                {
                    // unknown
                }
                else
                {
                    // axiom
                    ls->mAxiom[axiomcount*2+0] = l.tokendup(0);
                    ls->mAxiom[axiomcount*2+1] = l.tokendup(1);
                    axiomcount++;
                }
            }
        }
    }
    return ls;
}

